/*
 * Medical Image Registration ToolKit (MIRTK)
 *
 * Copyright 2013-2015 Imperial College London
 * Copyright 2013-2015 Andreas Schuh
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef MIRTK_NumericsConfig_H
#define MIRTK_NumericsConfig_H


/// Whether MIRTK Numerics module was built with VTK
#define MIRTK_Numerics_WITH_VTK 1

/// Whether MIRTK Numerics module was built with MATLAB
#define MIRTK_Numerics_WITH_MATLAB 0

/// Whether MIRTK Numerics module was built with ARPACK
#define MIRTK_Numerics_WITH_ARPACK 0

/// Whether MIRTK Numerics module was built with UMFPACK
#define MIRTK_Numerics_WITH_UMFPACK 0

/// Whether MIRTK Numerics module was built with LBFGS optimizer
#define MIRTK_Numerics_WITH_LBFGS 0

/// Whether SparseMatrix::Eigenvalues and SparseMatrix::Eigenvectors
/// functions are functional because eigs implementation was built
#define MIRTK_Numerics_WITH_eigs \
  MIRTK_Numerics_WITH_ARPACK && MIRTK_Numerics_WITH_UMFPACK


namespace mirtk {


/// Initialize MIRTK Numerics library
///
/// Registers all available types with their respective object factories.
void InitializeNumericsLibrary();


} // namespace mirtk

#endif // MIRTK_NumericsConfig_H
