##############################################################################
# @file  MIRTKTransformationConfig.cmake
# @brief CMake package configuration file of Transformation of MIRTK.
#
# @note The MIRTKTransformationConfig.cmake file is automatically generated
#       by BASIS from the template file ModuleConfig.cmake.in which is part of the
#       BASIS installation.
#
# @ingroup BasisConfig
##############################################################################

# ============================================================================
# namespace
# ============================================================================

# prefix used for variable names
set (NS "MIRTK_Transformation_")

# allow caller to change namespace - used by projects with modules
if (${NS}CONFIG_PREFIX)
  set (NS "${${NS}CONFIG_PREFIX}_")
endif ()

# ----------------------------------------------------------------------------
macro (basis_set_config NAME)
  set (${NS}${NAME} ${ARGN})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_copy_config NAME SRC)
  set (${NS}${NAME} ${${NS}${SRC}})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_append_config NAME)
  list (APPEND ${NS}${NAME} ${ARGN})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_remove_duplicates_config NAME)
  if (${NS}${NAME})
    list (REMOVE_DUPLICATES ${NS}${NAME})
  endif ()
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_remove_item_config NAME ITEM)
  if (${NS}${NAME})
    list (REMOVE_ITEM ${NS}${NAME} "${ITEM}")
  endif ()
endmacro ()

# ============================================================================
# general
# ============================================================================

## @brief Installation prefix or root of build tree.
basis_set_config (INSTALL_PREFIX "${CMAKE_CURRENT_LIST_DIR}/../../../")
get_filename_component (${NS}INSTALL_PREFIX "${${NS}INSTALL_PREFIX}" ABSOLUTE)

## @brief Project use file to be included by other projects.
basis_set_config (USE_FILE "${CMAKE_CURRENT_LIST_DIR}/MIRTKTransformationUse.cmake")

## @brief Directory of example files.
basis_set_config (EXAMPLE_DIR "")

# ============================================================================
# libraries
# ============================================================================

## @brief Directory of header files.
basis_set_config (INCLUDE_DIR "${${NS}INSTALL_PREFIX}/include")

## @brief Directory of header files including dependencies.
basis_set_config    (INCLUDE_DIRS) # to have a set recognized by Doxygen filter
basis_copy_config   (INCLUDE_DIRS INCLUDE_DIR)
basis_append_config (INCLUDE_DIRS "")

## @brief Directory of libraries.
basis_set_config (LIBRARY_DIR "${${NS}INSTALL_PREFIX}/lib/mirtk")

## @brief Directory of libraries including dependencies.
basis_set_config    (LIBRARY_DIRS) # to have a set() recognized by Doxygen filter
basis_copy_config   (LIBRARY_DIRS LIBRARY_DIR)
basis_append_config (LIBRARY_DIRS "")

## @brief Search path for Python modules.
basis_set_config (PYTHONPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/python")

## @brief Search path for Jython modules.
basis_set_config (JYTHONPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/jython")

## @brief Search path for Perl modules.
basis_set_config (PERL5LIB "${${NS}INSTALL_PREFIX}/lib/mirtk/perl")

## @brief Search path for MATLAB modules.
basis_set_config (MATLABPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/matlab")

## @brief Search path for Bash modules.
#
# See the import() function of the BASIS Bash utilities.
basis_set_config (BASHPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/bash")

# ============================================================================
# exports
# ============================================================================

## @brief Absolute path of exports file to be included by other projects.
basis_set_config (EXPORTS_FILE "${CMAKE_CURRENT_LIST_DIR}/MIRTKTransformationTargets.cmake")

## @brief Absolute path of custom exports file to be included by other projects.
basis_set_config (CUSTOM_EXPORTS_FILE "${CMAKE_CURRENT_LIST_DIR}/MIRTKTransformationCustomTargets.cmake")

# ============================================================================
# dependencies
# ============================================================================

set (_depwarn " set to different value than during the configuration of Transformation."
              " Using different versions of a dependency may cause inconsistencies!")
# Boost
if (DEPENDS_Boost_DIR)
  if (NOT DEPENDS_Boost_DIR STREQUAL "/usr/local")
    message (WARNING DEPENDS_Boost_DIR ${_depwarn})
  endif ()
else ()
  basis_set_or_update_value (DEPENDS_Boost_DIR "/usr/local")
endif ()
# TBB
if (DEPENDS_TBB_DIR)
  if (NOT DEPENDS_TBB_DIR STREQUAL "/usr/local")
    message (WARNING DEPENDS_TBB_DIR ${_depwarn})
  endif ()
else ()
  basis_set_or_update_value (DEPENDS_TBB_DIR "/usr/local")
endif ()
# VTK
if (DEPENDS_VTK_DIR)
  if (NOT DEPENDS_VTK_DIR STREQUAL "/usr/local/lib/cmake/vtk-9.2")
    message (WARNING DEPENDS_VTK_DIR ${_depwarn})
  endif ()
else ()
  basis_set_or_update_value (DEPENDS_VTK_DIR "/usr/local/lib/cmake/vtk-9.2")
endif ()
unset (_depwarn)
